import sympy
from sympy import *
import numpy as np


class BFGS:
    def find(z, Xl, Xu, Yl, Yu, point_for_start, e, iter):
        """
        Функция находит экстремум функции с помощью Алгоритма Бройдена — Флетчера — Гольдфарба — Шанно.
        Parameters
        ===========
        :param z: str
            строковая функция f(x)
        :param Xl: float
            интервал "от" по х
        :param Xu: float
            интервал "до" по х
        :param Yl: float
            интервал "от" по y
        :param Yu: float
            интервал "до" по y
        :param point_for_start: float
            точка старта
        :param e: float
            эпсилон - точность исследования
        :param flag: int
            если 1 то промежуточные расчеты выводятся если 0 то нет
        :param iter: int
            количество итераций
        Returns
        ===========
        экстремум функции
        """

        func = sympify(z)
        x, y, alpha = symbols('x y alpha')
        grad = [func.diff(x), func.diff(y)]

        i = 0
        while i < iter:
            grad_xk = np.array([[grad[0].subs([(x, point_for_start[0][0]), (y, point_for_start[1][0])])],
                                [grad[1].subs([(x, point_for_start[0][0]), (y, point_for_start[1][0])])]])

            sqrt_func = (grad_xk[0][0] ** 2 + grad_xk[1][0] ** 2) ** (1 / 2)

            if abs(sqrt_func) > e:
                I = np.array([[1, 0], [0, 1]])
                if i == 0:
                    H0 = I
                else:
                    I = H1

                p0 = -I.dot(grad_xk)

                matrix_with_alpha = point_for_start + alpha * p0

                f = func.subs([(x, matrix_with_alpha[0][0]), (y, matrix_with_alpha[1][0])])
                a = float(solve((f.diff(alpha), alpha)[0])[0])
                point_next = point_for_start + a * p0
                s0 = point_next - point_for_start

                grad_xk_1 = np.array([[grad[0].subs([(x, point_next[0][0]), (y, point_next[1][0])])],
                                      [grad[1].subs([(x, point_next[0][0]), (y, point_next[1][0])])]])

                y0 = grad_xk_1 - grad_xk

                k = 1 / np.dot(y0.transpose(), s0)

                A1 = I - k * (s0 * y0.transpose())
                A2 = I - k * (y0 * s0.transpose())

                H1 = np.dot(A1, np.dot(H0, A2)) + k * np.dot(s0, s0.transpose())

                point_for_start = point_next
                H0 = H1


            else:
                if (Xl <= point_for_start[0][0] <= Xu) and (Yl <= point_for_start[1][0] <= Yu):
                    print((point_for_start[0][0], point_for_start[1][0]))
                else:
                    print("Экстремум min не найден")
                break

            i += 1



import numpy as np
import sympy
from sympy import *
import time
from .GoldenSection import *
import matplotlib.pyplot as plt
import numexpr as ne
import matplotlib.animation as animation


class PorabolMet:
    def find(z, Xl, Xu, e, flag, iter):
        """
        Функция находит экстремум функции одной переменной методом парабол.
        Parameters
        ===========
        :param z: str
            строковая функция
        :param Xl: float
            интервал "от" по х
        :param Xu: float
            интервал "до" по х
        :param e: float
            эпсилон - точность исследования
        :param flag: int
            если 1 то промежуточные расчеты выводятся если 0 то нет
        :param iter: int
            количество итерация
        Returns
        ===========
        экстремум функции и график с движением точек
        """
        x1, x3 = Xl, Xu
        func = sympify(z)
        x = Symbol('x')
        x2 = GoldenSection()
        x2.find(z, x1, x3, e, 500, 'False', iter)
        x2 = x2.x_min
        lst_x_ = []
        i = 0
        if flag == 1:
            print(f'промежуточное значение х:')
        while i <= iter:
            if x1 < x2 < x3:
                func1 = float(func.subs(x, x1))
                func2 = float(func.subs(x, x2))
                func3 = float(func.subs(x, x3))

                if (func1 >= func2 <= func3 and i == 0) or i != 0:
                    a1 = (func2 - func1) / (x2 - x1)
                    a2 = (1 / (x3 - x2)) * (((func3 - func1) / (x3 - x1)) - ((func2 - func1) / (x2 - x1)))
                    x_ = (1 / 2) * (x1 + x2 - a1 / a2)
                    lst_x_.append(x_)
                    if flag == 1:
                        print(x_)
                    if x_ > x2:
                        x1 = x2
                        x2 = x_
                    elif x1 < x_ < x2:
                        x1 = x_

                    if i > 0:
                        if abs(lst_x_[i - 1] - lst_x_[i]) < e:
                            x_min = x_
                            print(f'Минимальное значение: {x_min:>.3f}')
                            break

            i += 1

        fig, ax = plt.subplots()

        ax.set_title('Поиск экстремумов')

        x = np.linspace(Xl, Xu, 50)
        y = ne.evaluate(z)

        plt.plot(x, y, color='b')

        redDot, = plt.plot([], [], 'ro')
        blueDot, = plt.plot([], [], 'bo')

        def animate(x):
            redDot.set_data(x, ne.evaluate(z))
            blueDot.set_data(x, ne.evaluate(z))

            return redDot, blueDot,

        # create animation using the animate() function
        myAnimation = animation.FuncAnimation(fig, animate, frames=np.linspace([Xl, Xu], x_min, 100),
                                              interval=10, blit=True, repeat=False)

        plt.show()
# print("Введите функцию f(x, y). Например:  -5*x**5 + 4*x**4 - 12* x**3 + 11*x**2 - 2*x - 1")
# f = "-3*x*sin(0.75*x) + exp(-2*x)"
# print("Введите левое ограничение по x. Например: 0")
# Xl = 0
# print("Введите правое ограничение по x. Например: 6")
# Xu = 6
# print("Введите e. Например: 0.001")
# e = 0.001
# ea = 100
# flag = 'False'
# print("Введите кол-во итераций. Например: 10")
# iter = 500
# functions = PorabolMet()
# functions.find(f, Xl, Xu, e, flag, iter)



import sympy
from sympy import *
import time
import math


class BrentMet:

    def PoiskVershin(x1, y1, x2, y2, x3, y3):
        denom = (x1 - x2) * (x1 - x3) * (x2 - x3)
        a1 = (x3 * (y2 - y1) + x2 * (y1 - y3) + x1 * (y3 - y2)) / denom
        b1 = (x3 * x3 * (y1 - y2) + x2 * x2 * (y3 - y1) + x1 * x1 * (y2 - y3)) / denom
        c1 = (x2 * x3 * (x2 - x3) * y1 + x3 * x1 * (x3 - x1) * y2 + x1 * x2 * (x1 - x2) * y3) / denom
        xv = -b1 / (2 * a1)
        # yv = c1 - b1 * b1 / (4 * a1)
        return xv

    def find(y, a, b, eps, flag, iterations):
        """
        Функция находит экстремум функции одной переменной методом Брента.
        Parameters
        ===========
        :param y: str
            строковая функция f(x)
        :param a: float
            интервал "от" по х
        :param b: float
            интервал "до" по х
        :param eps: float
            эпсилон - точность исследования
        :param flag: int
            если 1 то промежуточные расчеты выводятся если 0 то нет
        :param iterations: int
            количество итераций
        Returns
        ===========
        экстремум функции, если flag=1, выводятся промежуточные значения
        """
        func = sympify(y)
        x = Symbol('x')
        r = (3 - 5 ** (1 / 2)) / 2
        xx = a + r * (b - a)
        w = a + r * (b - a)
        v = a + r * (b - a)
        d_cur = b - a
        d_prv = b - a
        i = 0
        ii = 0
        while i < iterations:
            i += 1
            #print(abs(xx - a), abs(b - xx), max(abs(xx - a), abs(b - xx)))
            if max(abs(xx - a), abs(b - xx)) < eps:
                ii += 1
                print(f'Найденное решение: {float(xx):>.4f} за {i} шагов')
                break
            g = d_prv / 2
            d_prv = d_cur
            function1 = BrentMet()
            u = function1.PoiskVershin(xx, func.subs(x, xx), w, func.subs(x, w), v, func.subs(x, v))
            # print(f'{float(u):>.8f}')
            if u == nan:
                if xx < (a + b) / 2:
                    u = xx + r * (b - xx)
                    d_prv = b - xx
                else:
                    u = xx - r * (xx - a)
                    d_prv = xx - a

            elif ((a > u) and (u > b)) or (abs(u - xx) > g):
                if xx < (a + b) / 2:
                    u = xx + r * (b - xx)
                    d_prv = b - xx
                else:
                    u = xx - r * (xx - a)
                    d_prv = xx - a

            d_cur = abs(u - xx)
            #x33 = func.subs(x, u)
            # print(simplify(sympify(x33)))
            # print(float(func.subs(x, u)), 111)
            if float(func.subs(x, u)) > float(func.subs(x, xx)):
                if u > xx:
                    a = u
                else:
                    b = u
                if float(func.subs(x, u)) <= float(func.subs(x, w)) or (w == xx):
                    v = w
                    w = u
                elif float(func.subs(x, u)) <= float(func.subs(x, v)) or v == xx or v == w:
                    v = u
            else:
                if u < xx:
                    b = xx
                else:
                    a = xx
                v = w
                w = xx
                xx = u
            if flag == 1:
                print(f'итерация {i} минимальные значение {xx}')
        if ii != 1:
            print(f'Найдено неточное значение: {xx}')
            #function3 = BrentMet()
            #function3.find(y, int(xx)-a/3, int(xx)+b/3, 0.001, 100)



#function = BrentMet()
#function.find('4 - x ** 2 - 0.2 * x ** 3', -4, -2, 0.001, 500)
